import typer
import configparser
import requests
import time
import os
import webbrowser

app = typer.Typer()

AUTH0_DOMAIN = 'positron-beta.us.auth0.com'
AUTH0_CLIENT_ID = 'ZW0vio95rYfbHrN7kE3PoUXwmPloBw7e'
AUTH0_AUDIENCE = 'https://beta/positron/api'
ALGORITHMS = ['RS256']

POSITRON_API_BASE = 'https://beta.positronsupercompute.com/backend/api'
LOCAL_API_BASE = 'http://localhost:3000/api'

@app.command()
def hello():
    print('Hello, I am the Positron CLI')
    print('Here is a list of thing I can help you with:')
    print('- Logging you in to your Positron account and create your configuration file that will help you run jobs easily in the cloud')

@app.command()
def login(local: bool = False):
    if local:
        global POSITRON_API_BASE; POSITRON_API_BASE = LOCAL_API_BASE

    # Get device code
    print('Requesting device code')
    device_code_payload = {
        'client_id': AUTH0_CLIENT_ID,
        'scope': 'openid profile',
        'audience': AUTH0_AUDIENCE
    }
    device_code_response = requests.post(f'https://{AUTH0_DOMAIN}/oauth/device/code', data=device_code_payload)
    if device_code_response.status_code != 200:
        print('Error generating device code')
        raise typer.Exit(code=1)
    
    device_code_data = device_code_response.json()
    
    # Redirect to login
    print('1. On your computer or mobile device navigate to: ', device_code_data['verification_uri_complete'])
    print('2. Enter the following code: ', device_code_data['user_code'])
    print('3. Complete the login process')
    print('')
    webbrowser.open(url=device_code_data['verification_uri_complete'], new=2, autoraise=True)

    # Wait for authentication
    access_token = wait_for_access_token(device_code_data)
    
    # Get user token
    print('Requesting User Auth Token')
    auth_header = {
        'Authorization': f'Bearer {access_token}'
    }
    user_token_response = requests.get(f'{POSITRON_API_BASE}/get-user-auth-token', headers=auth_header)

    if user_token_response.status_code != 200:
        print(user_token_response.json())
        print('Error getting User Auth Token')
        raise typer.Exit(code=1)
    user_token_response_data = user_token_response.json()

    # Save user token
    print(f'Creating positron configuration at: {os.path.expanduser("~/.positron")}')
    save_user_token(user_token_response_data['userAuthToken'])


def wait_for_access_token(device_code_data):
    token_payload = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': device_code_data['device_code'],
        'client_id': AUTH0_CLIENT_ID
    }
    authenticated = False
    while not authenticated:
        print('Checking if the user completed the flow...')
        token_response = requests.post(f'https://{AUTH0_DOMAIN}/oauth/token', data=token_payload)

        token_data = token_response.json()
        if token_response.status_code == 200:
            print('Authenticated!')
            authenticated = True
            return token_data['access_token']
        elif token_data['error'] not in ('authorization_pending', 'slow_down'):
            print(token_data['error_description'])
            raise typer.Exit(code=1)
        else:
            time.sleep(device_code_data['interval'])

def save_user_token(user_token):
    try:
        config_folder = os.path.expanduser('~/.positron')
        if not os.path.exists(config_folder):
            os.mkdir(config_folder)
        config_path = os.path.expanduser('~/.positron/config.ini')
        config = configparser.ConfigParser()
        config['DEFAULT'] = {'UserAuthToken': user_token}
        with open(config_path, 'w') as config_file:
            config.write(config_file)
    except:
        print('Could not create configuration file')
        typer.Exit(code=1)

if __name__ == "__main__":
    app()


    


from functools import wraps
import os
import requests
import tarfile
import asyncio
import socketio
from colorama import init, Fore
import time
import configparser
import argparse
import signal
import json
import yaml
import sys

# Initialize colorama
init(autoreset=True)

# Init configuration
config = configparser.ConfigParser()

# Parse command line arguments
parser = argparse.ArgumentParser(description = "A decorator to handle deploying your code into the cloud")
parser.add_argument('--positron-deploy', action='store_true', help='Deploy your script into Positron Cloud', dest='deploy')
parser.add_argument('--stream-stdout', action='store_true', help='Stream the stdout from Positron Cloud back to your cli', dest='stream_stdout')
parser.add_argument('--debug', action='store_true', help='Get more detailed error messages', dest='debug')
positron_args, job_args = parser.parse_known_args()

# Cross component enums
log_types = dict(
    stdout="INFO",
    stderr="ERROR",
    debug="DEBUG"
)

# Decorator definition
def positron_sync(**positron_parameters):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if positron_args.deploy:
                positron_deploy(positron_parameters)
            else:
                parser.print_help()
                func(*args, **kwargs)

        return wrapper
    return decorator

# Cloud deployment definition
def positron_deploy(positron_parameters):
    signal.signal(signal.SIGINT, handle_sigint)
    global API_BASE
    global API_CREATE_JOB
    global API_GET_PRESIGNED
    global API_START_JOB
    global SOCKET_IO_DOMAIN, SOCKET_IO_PATH
    global COMPRESSED_WS_NAME

    API_BASE = positron_parameters.get('dev_api', 'https://beta.positronsupercompute.com/backend/api')
    SOCKET_IO_DOMAIN = positron_parameters.get('dev_ws', 'wss://beta.positronsupercompute.com')
    SOCKET_IO_PATH = '/backend/api/ws/socket.io' # '/api/ws/socket.io' # 
    API_CREATE_JOB = f'{API_BASE}/create-job'
    API_GET_PRESIGNED = f'{API_BASE}/generate-presigned-url'
    API_START_JOB = f'{API_BASE}/start-job'
    COMPRESSED_WS_NAME = 'workspace.tar.gz'

    try:
        print('Loading Configuration')
        load_config()
        load_job_config(positron_parameters)

        print('Validating sync parameters')
        validate_sync_parameters(parameters=positron_parameters)

        global auth_token, funding_group, image_name, environment_id, workspace_dir, entry_point
        auth_token = config['DEFAULT']['UserAuthToken']
        funding_group = positron_parameters.get('funding_group')
        image_name = positron_parameters.get('image_name')
        environment_id = positron_parameters.get('environment_id')
        workspace_dir = positron_parameters.get('workspace_dir')
        entry_point = positron_parameters.get('entry_point')

        print('Creating workspace tar file')
        create_workspace_tar()

        print('Creating new job')
        job = create_job()

        print('Fetching presigned url for upload')
        resp = get_presigned_url(job_id=job['id'])

        print('Uploading compressed workspace to Positron storage')
        upload_file(resp.get('url'), resp.get('fields'))

        print('Starting Job')
        start_job(job_id=job['id'])

        print(f"Your workspace has been uploaded and the job is in a processing queue. Please check your dashboard to follow your jobs {job['name']} status!")

        start_stdout_stream(job['id'])
        
    except PositronException as e:
        print(e)
    except Exception as e:
        print('An exception occured. Please use --debug flag to get details')
        debug(e)


def load_config():
    env_token = os.getenv('PositronUserAuthToken', False)
    config_path = os.path.expanduser('~/.positron/config.ini')
    if env_token:
        config['DEFAULT'] = {'UserAuthToken': env_token}
    elif os.path.exists(config_path):
        config.read(config_path)    
    else:
        raise PositronException('Missing UserAuthToken, please init configuration')


def load_job_config(parameters):
    script_dir = os.path.dirname(sys.argv[0])
    config_path = os.path.join(script_dir, "job_config.yaml")
    if not os.path.exists(config_path):
        return
    try:
        with open(config_path, 'r') as job_config_file:
            job_config = yaml.safe_load(job_config_file)
            keys = [
                'funding_group',
                'environment_id',
                'image_name'
            ]
            for key in keys:
                config = job_config['python_job'][key]
                if not parameters.get(key, False) and config is not None:
                    parameters[key] = config
    except Exception as e:
        debug(f'Error loading job configuration! {str(e)}')


def validate_sync_parameters(parameters):
    valid = True
    if not parameters.get('funding_group', False):
        print('Missing decorator parameter: funding_group')
        valid = False
    if not parameters.get('image_name', False):
        print('Missing decorator parameter: image_name')
        valid = False
    if not parameters.get('environment_id', False):
        print('Missing decorator parameter: environment_id')
        valid = False
    if not parameters.get('workspace_dir', False):
        print('Missing decorator parameter: workspace_dir')
        valid = False
    if not parameters.get('entry_point', False):
        print('Missing decorator parameter: entry_point')
        valid = False

    if not valid:
        raise PositronException(
            'Validation failed! Please check your decorator parameters!')


def create_workspace_tar():
    # Use the context manager to handle opening and closing the tar file
    with tarfile.open(COMPRESSED_WS_NAME, 'w:gz') as tar:
        for root, dirs, files in os.walk(workspace_dir):
            for exclude_dir in ['.venv', '.git', '__pycache__']:
                try:
                    dirs.remove(exclude_dir)
                except ValueError:
                    # Ignore if the directory is not in the list
                    pass
            for file in files:
                full_path = os.path.join(root, file)
                arcname = os.path.relpath(full_path, start=workspace_dir)
                tar.add(full_path, arcname=arcname)


def create_job():
    debug(f'Calling: {API_CREATE_JOB}')
    data = {"imageName": image_name, "fundingGroupId": funding_group, "environmentId": environment_id, "entryPoint": entry_point, "jobArguments": job_args}
    debug(data)
    Headers = {"PositronAuthToken": auth_token}
    response = requests.post(API_CREATE_JOB, headers=Headers, json=data)
    debug(response)
    if response.status_code != 200:
        raise PositronException(
            f'Job creation failed with http code: {response.status_code} \n {response.text}')
    else:
        debug(response.json())
        return response.json()


def get_presigned_url(job_id):
    Headers = {"PositronAuthToken": auth_token, "PositronJobId": job_id}
    url = API_GET_PRESIGNED + '?filename=' + COMPRESSED_WS_NAME
    debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    debug(response)
    if response.status_code != 200:
        raise PositronException(
            f'Presigned url fetching failed with http code: {response.status_code} \n {response.text}')
    else:
        debug(response.json())
        return response.json()


def upload_file(url, data):
    with open(COMPRESSED_WS_NAME, 'rb') as f:
        files = {'file': f}
        response = requests.post(url, data=data, files=files)
        if response.status_code != 204:
            raise PositronException(
                f'Upload failed with http code: {response.status_code} \n {response.text}')


def start_job(job_id):
    Headers = {"PositronAuthToken": auth_token, "PositronJobId": job_id}
    debug(f'Calling: {API_START_JOB}')
    response = requests.get(API_START_JOB, headers=Headers)
    debug(response)
    if response.status_code != 200:
        raise PositronException(
            f'Failed to start job with http code: {response.status_code} \n {response.text}')
    else:
        debug(response.json())
        return response.json()


def start_stdout_stream(job_id):
    if positron_args.stream_stdout:
        try:
            asyncio.get_event_loop().run_until_complete(start_stream(job_id))
        except Exception as e:
            print (e)


sio = socketio.AsyncClient()


@sio.event(namespace='/stdout-stream')
async def connect():
    print('Connected to stdout stream')


@sio.event(namespace='/stdout-stream')
async def message(message):
    try:
        log = json.loads(message)
        if log['log_level'] == log_types['stdout']:
            print(Fore.GREEN + log['message'])
        elif log['log_level'] == log_types['stderr']:
            print(Fore.RED + log['message'])
        elif log['log_level'] == log_types['debug']:
            print(Fore.BLUE + log['message'])            
    except:
        print(message)


@sio.event(namespace='/stdout-stream')
async def disconnect():
    print('Disconnected from stdout stream')


@sio.event(namespace='/stdout-stream')
async def error(err):
    print('An error occured in the streaming process')
    debug(err)


async def start_stream(job_id):
    custom_headers = {
        "PositronAuthToken": auth_token,
        "PositronJobId": job_id
    }
    await sio.connect(SOCKET_IO_DOMAIN, headers=custom_headers, socketio_path=SOCKET_IO_PATH)
    await sio.wait()


def handle_sigint(signum, frame):
    print('Terminating gracefully...')
    time.sleep(5)
    exit(0)

class PositronException(Exception):
    "Raised when decorator related errors are happening"
    pass


def debug(log):
    if positron_args.debug:
        print(log)



from setuptools import setup, find_packages

setup(
    name="positron_networks",
    version="0.1.10",
    description="Run experiments on the Positron Cloud",
    author="Balint Kerdi",
    author_email="bkerdi@positronnetworks.com",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(where="app"),
    package_dir={"": "app"},
    install_requires=[
        'requests>=2.20.0,<3.0.0',
        'python-socketio>=5.0.0,<6.0.0',
        'colorama>=0.4.0,<1.0.0',
        'configparser>=5.0.0,<7.0.0',
        'argparse>=1.4.0',
        'aiohttp>=3.7.4.post0',
        'typer>=0.12.3',
        'pyyaml>=6.0.1'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    license='Apache License 2.0',
)
